"""
Rohan Sanda, Akhilesh Balasingam 2024

This file contains the wrappers for running different methods on the BSDS300 dataset.
Set configuration parameters in config.py file.
You may have to modify this if you change the dataset or the methods you want to run.
"""

import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
from glob import glob
import os
from skimage.metrics import peak_signal_noise_ratio as compare_psnr
import skimage.io
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from tqdm import tqdm
import pickle
from config import *
from utils import *

from config import *
from models_dip import *
from utils import *
from denoise_deblur_dip import *
from hw5_task2 import run_hw5
from denoise_dip_tv import admm_dip_single
from bm3d_wrapper import bm3d_single
from hw5_task2 import BSDS300Dataset
from hw5_task3 import *
#from deblur_dip import smart_dip_deblur
import torchvision.transforms.functional as TF
# set random seeds
torch.manual_seed(1)
np.random.seed(1)
torch.use_deterministic_algorithms(True)
matplotlib.rcParams['figure.raise_window'] = False
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

methods_denoise = ["DCNN", "BM3D", "DIP", "ADMM-DIP"]
methods_hw_deblur = ["weiner", "deblur_denoise", "denoise"]
methods_deblur = ["task3", "smart_dip_deblur", "dip", "admm_dip"]
                
def run_method(dataset, dataset_name="BSDS300", task="denoise", method= "DIP", fsavepath="./results", verbose=False, sigma=0.01):
        if not os.path.exists(fsavepath):
            os.makedirs(fsavepath)
            
        print(f"Running {method} on {dataset_name} for {task}")
            
        if dataset_name == "BSDS300":   
            subset = get_random_subset(dataset, subset_size=5)
            
            processed_pils, processed_tensors = process_subset(subset)
            n = len(processed_pils)
            if task == "denoise":
                if not os.path.exists(fsavepath + "/denoise/" + method):
                    os.makedirs(fsavepath + "/denoise/" + method)
                    
                if method == "DCNN":
                    metrics, model = run_hw5(subset)
                    with open(fsavepath + "/denoise/" + method + "/SPECKLE_model_sigma=0.10.pkl", "wb") as f:
                        pickle.dump(model, f)
                        
                    with open(fsavepath + "/denoise/" + method + "/SPECKLE_metrics_sigma=0.10.pkl", "wb") as f:
                        pickle.dump(metrics, f)
                
                else:
                    for i in range(n):
                        img_pil = processed_pils[i]
                        img_tensor = processed_tensors[i]
                        img_np = processed_tensors[i].numpy()
                        
                        sigma = ALL_PARAMS[method]["sigma"]
                        img_noisy_np = add_noise(img_tensor, sigma).numpy()
                        metrics_all = []
                        
                        if method == "BM3D":
                            metrics, output = bm3d_single(img_pil, img_np, img_noisy_np, i+1, verbose=verbose)
                        elif method == "DIP":
                            metrics, output = dip_single(img_pil, img_np, img_noisy_np, i+1, verbose=verbose)
                            with open(fsavepath + "/denoise/" + method + f"/model_image={i+1}.pkl", "wb") as f:
                                pickle.dump(output, f)
                        elif method == "ADMM-DIP":
                            metrics, output = admm_dip_single(img_pil, img_np, img_noisy_np, i+1, verbose=verbose)
                            with open(fsavepath + "/denoise/" + method + f"/model_image={i+1}.pkl", "wb") as f:
                                pickle.dump(output, f)
                        else:
                            assert False
                        
                        metrics_all.append(metrics)
                        
                        # Save metrics
                        os.makedirs(fsavepath + "/denoise/" + method, exist_ok=True)
                        with open(fsavepath + "/denoise/" + method + f"/metrics_image={i+1}.pkl", "wb") as f:
                            pickle.dump(metrics, f)
            elif task == "deblur":
                dataset = BlurredBSDS300Dataset(split='test', kernel_size=7)
                subset = dataset
                idx=0
                for img, gt, kernel in subset:
                    gt_pil = TF.to_pil_image(gt.squeeze(0))
                    gt_pil.save(fsavepath + "/deblur/" + f"gt{idx}.png")
                    img = img + SMART_DEBLUR_DIP_PARAMS["sigma"]*torch.randn_like(img)
                    idx += 1
                if method in methods_hw_deblur:
                    method = "task3"
                    if not os.path.exists(fsavepath + "/deblur/" + method):
                        os.makedirs(fsavepath + "/deblur/" + method)  
                        
                    metrics = run_hw5_task3(subset, sigma_in=sigma)
                    print(metrics)
                    with open(fsavepath + "/deblur/" + method + f"/metrics_image_tot_hw5_task3.pkl", "wb") as f:
                        pickle.dump(metrics, f)
                    # with open(fsavepath + "/deblur/" + method + f"/out_dump_hw5_task3.pkl", "wb") as f:
                    #     pickle.dump(out_dump, f)
                elif method == "SDDP":
                    ind = 0
                    if not os.path.exists(fsavepath + "/deblur/" + method):
                        os.makedirs(fsavepath + "/deblur/" + method)  
                        
                    for img, gt, kernel in subset:
                        size1 = kernel.shape[-1]
                        img_np = img.squeeze(0).numpy()
                        gt_pil = TF.to_pil_image(gt.squeeze(0))
                        gt_np = gt.squeeze(0).numpy()
                        #metrics, out_dump = smart_dip_deblur(gt_pil, gt_np, img_np, size1, ind, fsave=fsavepath + "/deblur/" + method, verbose=verbose)
                        
                        os.makedirs(fsavepath + "/deblur/" + method, exist_ok=True)
                        with open(fsavepath + "/deblur/" + method + f"/metrics_image={ind+1}.pkl", "wb") as f:
                            pickle.dump(metrics, f)
                        
                        ind += 1
                elif method == "DIP":
                    idx = 0
                    for img, gt, kernel in subset:
                        idx += 1
                        size1 = kernel.shape[-1]
                        img_np = img.squeeze(0).numpy()
                        gt_pil = TF.to_pil_image(gt.squeeze(0))
                        gt_np = gt.squeeze(0).numpy()
                        metrics, output = dip_single(gt_pil, gt_np, img_np, idx, deblur=True, verbose=verbose)
                        with open(fsavepath + "/deblur/" + method + f"/model_image={idx}.pkl", "wb") as f:
                            pickle.dump(output, f)
                        with open(fsavepath + "/deblur/" + method + f"/metrics_image={idx}.pkl", "wb") as f:
                            pickle.dump(metrics, f)
                elif method == "all":
                    idx = 0
                    for img, gt, kernel in subset:
                        idx += 1
                        size1 = kernel.shape[-1]
                        img_np = img.squeeze(0).numpy()
                        gt_pil = TF.to_pil_image(gt.squeeze(0))
                        gt_np = gt.squeeze(0).numpy()
                        
                        for method in methods_deblur:
                            if method == "task3":
                                if not os.path.exists(fsavepath + "/deblur/" + method):
                                    os.makedirs(fsavepath + "/deblur/" + method)  
                                    
                                metrics, out_dump = run_hw5_task3(subset, sigma_in=sigma)
                                print(metrics)
                                with open(fsavepath + "/deblur/" + method + f"/metrics_image_tot_hw5_task3.pkl", "wb") as f:
                                    pickle.dump(metrics, f)
                                with open(fsavepath + "/deblur/" + method + f"/out_dump_hw5_task3.pkl", "wb") as f:
                                    pickle.dump(out_dump, f)
                            else:
                                if not os.path.exists(fsavepath + "/deblur/" + method):
                                    os.makedirs(fsavepath + "/deblur/" + method)  
                                if method == "SDDP":
                                    continue
                                if method == "DIP":
                                    metrics, output = dip_single(gt_pil, gt_np, img_np, idx, deblur=True, verbose=verbose)
                                    with open(fsavepath + "/deblur/" + method + f"/model_image={idx}.pkl", "wb") as f:
                                        pickle.dump(output, f)
                                    with open(fsavepath + "/deblur/" + method + f"/metrics_image={idx}.pkl", "wb") as f:
                                        pickle.dump(metrics, f)
                        
        print(f"Finished running {method} on {dataset_name} for {task}")
            
def run_all_on_dataset(dataset, dataset_name="BSDS300", task="denoise", fsavepath="./results"):
    if not os.path.exists(fsavepath):
        os.makedirs(fsavepath)
    
    if dataset_name == "BSDS300":   
        subset = get_random_subset(dataset, subset_size=5)
        processed_pils, processed_tensors = process_subset(subset)
        n = len(processed_pils)
        
        if task == "denoise":
            for method in methods_denoise:
                run_method(dataset, dataset_name, task, method, fsavepath)
                               
            
                        
   
        
            
        

    